--- ========================================================= ---
---  Setup Revision, Version, Date and add to AceLibrary
--- ========================================================= ---
local VERSION = tonumber(("$Revision: 44247 $"):match("%d+"))

Sphere = AceLibrary("AceAddon-2.0"):new("AceDB-2.0", "AceEvent-2.0", "AceConsole-2.0", "AceModuleCore-2.0", "FuBarPlugin-2.0", "AceHook-2.1")
local Sphere, self = Sphere, Sphere
Sphere.version = "r" .. VERSION
Sphere.revision = VERSION
Sphere.date = ("$Date: 2007-07-19 15:21:49 -0400 (Thu, 19 Jul 2007) $"):match("%d%d%d%d%-%d%d%-%d%d")

--- ========================================================= ---
---  Setup Locals
--- ========================================================= ---
local Waterfall = AceLibrary:HasInstance("Waterfall-1.0") and AceLibrary("Waterfall-1.0")
local Dewdrop = AceLibrary("Dewdrop-2.0")
local DogTag = AceLibrary("DogTag-1.0")

local localeTables = {}
function Sphere:L(name)
	if not localeTables[name] then
		localeTables[name] = setmetatable({}, {__index = function(self, key)
			self[key] = key
			return key
		end})
	end
	return localeTables[name]
end

local L = Sphere:L("Sphere")

local start = nil

local eventType = {
	health = "[PercentHP]",
	power = "[PercentMP]",
	threat = "[PercentThreat]",
	experience = "[PercentXP]",
	reputation = "[PercentRep]",
	none = "",
	custom = "",
}

local frameCount = 0
local frames = { }

--- ========================================================= ---
---  Setup FuBar Configuration
--- ========================================================= ---
Sphere.hasIcon = true
Sphere.hideWithoutStandby = true
Sphere.defaultPosition = "MINIMAP"
Sphere.defaultMinimapPosition = 200
Sphere.hasNoText = false
Sphere.hasNoColor = true
Sphere.cannotDetachTooltip = true
Sphere.independentProfile = true
Sphere.blizzardTooltip = true

--- ========================================================= ---
---  Setup Database
--- ========================================================= ---

Sphere:RegisterDB("SphereDB")
Sphere:RegisterDefaults('char', {
		settings = {
			locked = false,
			clamped = true,
			scale = 1.20,
			strata = "MEDIUM",
		},
		colors = {
			health = {0, 1, 0, 1},
			mana = {0, 0.25, 1, 1},
			rage = {1, 0, 0, 1},
			energy = {1, 1, 0, 1},
			none = {0, 0, 0, 1},
			custom = {1, 1, 1, 1},
			combat = {1, 0, 0, 1},
			experience = {0.75, 0, 0.75, 1},
			reputation = {0, 0.75, 0.75, 1},
			threat = {1, 0, 0, 1},
		},
		gaugeUnit = "player",
		gaugeType = "health",
		gaugeCustom = "[PercentHP]",
		gaugeMarks = "five",
		gaugeMarksDarken = true,
		animate = true,
		useCustom = false,

})

--- ========================================================= ---
---  Setup Validates for Options
--- ========================================================= ---
local strataValidate = {
	BACKGROUND = L["Background"],
	LOW = L["Low"],
	MEDIUM = L["Medium"],
	HIGH = L["High"],
	DIALOG = L["Dialog"],
	TOOLTIP = L["Tooltip"],
}

local strataValidateDesc = {}

Sphere_eventTypeValidate = {
	health = L["Health"],
	power = L["Power"],
	threat = L["Threat"],
	experience = L["Experience"],
	reputation = L["Reputation"],
	custom = L["Custom"],
	none = L["None"],
}

local eventTypeValidateDesc = {}

local eventUnitValidate = {
	player = L["Player"],
	target = L["Target"],
	focus = L["Focus"],
	pet = L["Pet"],
}

local eventUnitValidateDesc = {}

local gaugeMarksValidate = {
	none = L["None"],
	four = L["Four"],
	five = L["Five"],
	ten = L["Ten"],
	twenty = L["Twenty"],
}

local gaugeMarksValidateDesc = {}

--- ========================================================= ---
---  Setup Options
--- ========================================================= ---
Sphere.options = {
	type = 'group',
	handler = Sphere,
	args = {
		appearance = {
			name = L["Appearance"],
			desc = L["Change how the Sphere Appearance looks in general."],
			type = 'group',
			args = {
				colors = {
					name = L["Colors"],
					desc = L["Change the Colors that the Sphere uses."],
					type = 'group',
					args = {
						healthColor = {
							name = L["Health Colors"],
							desc = L["Set what color your health color is."],
							type = 'group',
							pass = true,
							get = function(value)
								return unpack(self.db.char.colors[value])
							end,
							set = "SetColor",
							args = {
								health = {
									name = L["Health"],
									desc = L["Color for your Health."],
									type = 'color',
									hasAlpha = false,
								},
							},
						},
						powerColor = {
							name = L["Power Colors"],
							desc = L["Set what color each power color is."],
							type = 'group',
							pass = true,
							get = function(value)
								return unpack(self.db.char.colors[value])
							end,
							set = "SetColor",
							args = {
								mana = {
									name = L["Mana"],
									desc = L["Color for your Mana."],
									type = 'color',
									hasAlpha = false,
								},
								rage = {
									name = L["Rage"],
									desc = L["Color for your Rage."],
									type = 'color',
									hasAlpha = false,
								},
								energy = {
									name = L["Energy"],
									desc = L["Color for your Energy."],
									type = 'color',
									hasAlpha = false,
								},
							},
						},
						otherColor = {
							name = L["Other Colors"],
							desc = L["Set what color you want these to be."],
							type = 'group',
							pass = true,
							get = function(value)
								return unpack(self.db.char.colors[value])
							end,
							set = "SetColor",
							args = {
								none = {
									name = L["None"],
									desc = L["Color for None."],
									type = 'color',
									hasAlpha = false,
								},
								custom = {
									name = L["Custom"],
									desc = L["Color for Custom."],
									type = 'color',
									hasAlpha = false,
								},
								combat = {
									name = L["Combat"],
									desc = L["Color for when in Combat."],
									type = 'color',
									hasAlpha = false,
								},
								experience = {
									name = L["Experience"],
									desc = L["Color for your Experience."],
									type = 'color',
									hasAlpha = false,
								},
								reputation = {
									name = L["Reputation"],
									desc = L["Color for your Reputation."],
									type = 'color',
									hasAlpha = false,
								},
								threat = {
									name = L["Threat"],
									desc = L["Color for your Threat."],
									type = 'color',
									hasAlpha = false,
								},
							},
						},
					},
				},
				size = {
					name = L["Size"],
					desc = L["Set how large the Sphere is."],
					type = 'range',
					min = 0.25,
					max = 3,
					step = 0.01,
					bigStep = 0.05,
					get = function()
						return self.db.char.settings.scale
					end,
					set = "SetScale"
				},
				strata = {
					name = L["Frame Strata"],
					desc = L["Choose the layer the unit frames are located on."],
					type = 'text',
					validate = strataValidate,
					validateDesc = strataValidateDesc,
					get = function()
						return self.db.char.settings.strata
					end,
					set = "SetFrameStrata",
					disabled = InCombatLockdown,
				},
				clamped = {
					name = L["Clamp frames to screen"],
					desc = L["Make it so that frames cannot be dragged off-screen."],
					type = 'toggle',
					get = function()
						return self.db.char.settings.clamped
					end,
					set = function() 
						self.db.char.settings.clamped = not(self.db.char.settings.clamped); 
						self:SetClampedToScreen();
					end,
					disabled = InCombatLockdown,
				},
			},
		},
		outergauge = {
			name = L["Outer Gauge"],
			desc = L["Change the Outer Gauge options."],
			type = 'group',
			args = {
				event = {
					name = L["Event Type"],
					desc = L["Choose which Event Type the Gauge monitors and shows."],
					type = 'text',
					validate = Sphere_eventTypeValidate,
					validateDesc = eventTypeValidateDesc,
					get = function()
						return self.db.char.gaugeType
					end,
					set = "SetGaugeType",
				},
				unit = {
					name = L["Event Unit"],
					desc = L["Choose which Unit the Event is monitoring and showing."],
					type = 'text',
					validate = eventUnitValidate,
					validateDesc = eventUnitValidateDesc,
					get = function()
						return self.db.char.gaugeUnit
					end,
					set = "SetGaugeUnit",
				},
				custom = {
					name = L["Custom Event"],
					desc = L["Create a Custom Event for the Gauge to monitor and show."],
					type = "text",
					get = function()
						return self.db.char.gaugeCustom
					end,
					set = "SetCustomGauge",
					usage = L["DogTag-1.0 tags. See http://www.wowace.com/wiki/DogTag-1.0"],
				},
				marks = {
					name = L["Marks Number"],
					desc = L["Choose how many Marks to show on the Gauge."],
					type = 'text',
					validate = gaugeMarksValidate,
					validateDesc = gaugeMarksValidateDesc,
					get = function()
						return self.db.char.gaugeMarks
					end,
					set = "SetGaugeMarks",
				},
				marksdarken = {
					name = L["Marks Darken"],
					desc = L["Choose whether or not to Darken the Marks."],
					type = 'toggle',
					get = function()
						return self.db.char.gaugeMarksDarken
					end,
					set = function() 
						self.db.char.gaugeMarksDarken = not(self.db.char.gaugeMarksDarken);
						self:SetGaugeMarksDarken();
					end,
				},
				animate = {
					name = L["Animate Gauge"],
					desc = L["Animate the Gauge (Glide)."],
					type = 'toggle',
					get = function()
						return self.db.char.animate
					end,
					set = function() 
						self.db.char.animate = not(self.db.char.animate);
					end,
				},
			},
		},
	}
}

--- ========================================================= ---
---  Setup Menu Request and Tooltip
--- ========================================================= ---
Sphere.OnMenuRequest = Sphere.options

function Sphere:OnTooltipUpdate()
	GameTooltip:AddLine("Sphere")
	GameTooltip:AddLine(" ")
	GameTooltip:AddDoubleLine(LOCKED .. ":", (self.db.char.settings.locked and YES or NO), 1, 1, 0, 1, 1, 1)
	GameTooltip:AddLine(" ")
	GameTooltip:AddLine(L["|cffffff00Shift-Click|r to lock/unlock Sphere"], 0.2, 1, 0.2)
	if Waterfall then
		GameTooltip:AddLine(L["|cffffff00Left-Click|r to change settings with a nice GUI configuration."], 0.2, 1, 0.2)
	end
	GameTooltip:AddLine(L["|cffffff00Right-Click|r to change settings with a drop-down menu."], 0.2, 1, 0.2)
end

--- ========================================================= ---
---  Setup Waterfall
--- ========================================================= ---
if Waterfall then
	Waterfall:Register("Sphere", 
						"aceOptions", Sphere.options,
						"title", L["Sphere Configuration"],
						"treeLevels", 5);
end

--- ========================================================= ---
---  Setup Modules
--- ========================================================= ---

local modulesWithFunction = setmetatable({}, {__index = function(self, key)
	local t = {}
	for name, module in Sphere:IterateModules() do
		if type(module[key]) == "function" and Sphere:IsModuleActive(module) then
			t[#t+1] = module
		end
	end
	self[key] = t
	
	return t
end})

function Sphere:OnModuleCreated(name, module)
	for k in pairs(modulesWithFunction) do
		modulesWithFunction[k] = nil
	end
end

function Sphere.modulePrototype:RegisterSphereOption(data)
	local args = Sphere.options.args
	local name = self.name
	if args[name] then
		local i = 1
		while args[name .. i] do
			i = i + 1
		end
		name = name .. i
	end
	args[name] = data
	if not data.handler then
		data.handler = self
	end
	if data.args then
		data.args.active = {
			type = 'toggle',
			name = L["Enable"],
			desc = L["Enable this module"],
			get = "IsModuleActive",
			set = "ToggleModuleActive",
			handler = Sphere,
			passValue = self,
		}
	end
end

--- ========================================================= ---
---  On Intitalize
--- ========================================================= ---
function Sphere:OnInitialize()
	self:RegisterChatCommand('/Sphere', {
		type = 'execute',
		func = function()
			self:OpenMenu()
		end
	})
	
	for k,v in pairs(self.db.profile) do
		for name, module in self:IterateModules() do
			if module.db then
				local db = module.db.profile
				if rawget(db, k) then
					db[k] = v
					self.db.profile[k] = nil
					break
				end
			end
		end
	end
	
	self:CreateFrames()
	self:SetGaugeMarks()
	self:SetGaugeMarksDarken()
	self:SetScale()
end

--- ========================================================= ---
---  On Enable
--- ========================================================= ---
function Sphere:OnEnable()
	self:ScheduleRepeatingEvent("UpdateOuterGauge", self.UpdateGauge, 0.05, self)
	SphereMainFrame:Show();
	self:SetClampedToScreen()
end

--- ========================================================= ---
---  On Disable
--- ========================================================= ---
function Sphere:OnDisable()
	self:UnregisterAllEvents();
	SphereMainFrame:Hide();
end

--- ========================================================= ---
---  On Module Enable
--- ========================================================= ---
function Sphere:OnModuleEnable()
	for k in pairs(modulesWithFunction) do
		modulesWithFunction[k] = nil
	end
end

--- ========================================================= ---
---  On Module Disable
--- ========================================================= ---
function Sphere:OnModuleDisable()
	for k in pairs(modulesWithFunction) do
		modulesWithFunction[k] = nil
	end
end

--- ========================================================= ---
---  On Click
--- ========================================================= ---
function Sphere:OnClick()
	local shift, alt, ctrl = IsShiftKeyDown(), IsAltKeyDown(), IsControlKeyDown()
	if shift and not alt and not ctrl then
		self.db.char.settings.locked = not self.db.char.settings.locked
		self:UpdateTooltip()
	else	
		if Waterfall then
			Waterfall:Open("Sphere")
		else
			self:Print(L["Waterfall-1.0 is required to access the GUI."])
		end
	end
end

local showWhenValidate = {
	never = L["Never"],
	notcombat = L["Out of combat"],
	always = L["Always"],
}

--- ========================================================= ---
---  Add Event Type
--- ========================================================= ---
function Sphere:AddEventType(func, key, value)
	func[key] = value
end

--- ========================================================= ---
---  Remove Event Type
--- ========================================================= ---
function Sphere:RemoveEventType(func, key)
	func[key] = nil
end

--- ========================================================= ---
---  Set Color
--- ========================================================= ---
function Sphere:SetColor(given_kind, r, g, b, a)
	local kind = given_kind
	local colors = self.db.char.colors[kind]
	
	if r then
		colors[1] = r
		colors[2] = g
		colors[3] = b
		colors[4] = a
	else
		r, g, b, a = unpack(colors)
	end
end

--- ========================================================= ---
---  Set Scale
--- ========================================================= ---
function Sphere:SetScale(value)
	if value then
		self.db.char.settings.scale = value
	else
		value = self.db.char.settings.scale
	end
	
	SphereMainFrame:SetScale(value)
end

--- ========================================================= ---
---  Set Gauge Type
--- ========================================================= ---
function Sphere:SetGaugeType(value)
	if value then
		self.db.char.gaugeType = value
		self.db.char.gaugeCustom = eventType[value]
	else
		value = self.db.char.gaugeType
	end
	
	if value == "health" or value == "power" or value == "threat" or value == "experience" or value == "reputation" or value == "custom" or value == "none" then
		 self.db.char.useCustom = false
	else
		self.db.char.useCustom = true
	end
end

--- ========================================================= ---
---  Set Gauge Unit
--- ========================================================= ---
function Sphere:SetGaugeUnit(value)
	if value then
		self.db.char.gaugeUnit = value
	else
		value = self.db.char.gaugeUnit
	end
end

--- ========================================================= ---
---  Set Custom Gauge
--- ========================================================= ---
function Sphere:SetCustomGauge(value)
	if value then
		if self.db.char.gaugeType == "custom" then
			self.db.char.gaugeCustom = value
		else
			self.db.char.gaugeCustom = eventType[self.db.char.gaugeType]
		end
	else
		value = self.db.char.gaugeCustom
	end
end

--- ========================================================= ---
--- Set Gauge Marks
--- ========================================================= ---

function Sphere:SetGaugeMarks(value)
	if value then
		self.db.char.gaugeMarks = value
	else
		value = self.db.char.gaugeMarks
	end
	
	if value == "none" then value = 0 end
	if value == "four" then value = 4 end
	if value == "five" then value = 5 end
	if value == "ten" then value = 10 end
	if value == "twenty" then value = 20 end
	
	if value ~= 0 then
		SphereOuterMarks:SetNormalTexture("Interface\\Addons\\Sphere\\art\\outermarks" .. value);
		SphereOuterMarks:Show();
	else
		SphereOuterMarks:Hide();
	end

end

--- ========================================================= ---
--- Set Gauge Darken Marks
--- ========================================================= ---
function Sphere:SetGaugeMarksDarken()
	value = self.db.char.gaugeMarksDarken

	if value then
		SphereOuterMarks:SetAlpha(1);
	else
		SphereOuterMarks:SetAlpha(0.5);
	end

end

--- ========================================================= ---
---  Set Frame Strata
--- ========================================================= ---
function Sphere:SetFrameStrata(value)
	if value then
		self.db.char.settings.strata = value
	else
		value = self.db.char.settings.strata
	end
	
	SphereMainFrame:SetFrameStrata(value)
end

--- ========================================================= ---
---  Set Frames Clamped To Screen
--- ========================================================= ---

function Sphere:SetClampedToScreen()
	value = self.db.char.settings.clamped
	
	for index=1, frameCount do
		getglobal(frames[index].frameName):SetClampedToScreen(value)
	end
end

--- ========================================================= ---
--- Add To Frames
--- ========================================================= ---
function Sphere:AddToFrames(frameName)
	frameCount = frameCount + 1
	frames[frameCount] = {};
	frames[frameCount].frameName = frameName
end

--- ========================================================= ---
--- Create Frame
--- ========================================================= ---
function Sphere:CreateFrame(frameType, frameName, frameParent, frameLevel, textureFile, hasEvents)

	local f = CreateFrame(frameType, frameName, frameParent);
	f:SetID(0);
	f:SetWidth(64);
	f:SetHeight(64);
	f:SetPoint("Center");
	f:EnableMouse(true);
	
	if frameLevel == "top" then
		f:SetToplevel("true");
	else
		f:SetFrameLevel(frameLevel);
	end
	
	if frameType == "Button" then
		f:SetNormalTexture(textureFile);
	end
	
	if hasEvents then
		f:EnableMouse(true);
		f:SetMovable(true);
		f:RegisterForDrag("LeftButton");
	
		f:SetScript("OnUpdate", function() 
		
			end);
			
		f:SetScript("OnDragStart", function()

				if not(Sphere.db.char.settings.locked) then
					this:StartMoving(); 
				end
			end);
			
		f:SetScript("OnDragStop", function () 
			
				this:StopMovingOrSizing(); 
			
			end);
			
		f:SetScript("OnMouseUp", function ()

			end);

		f:SetScript("OnReceiveDrag",  function ()

			end);
	else
		f:EnableMouse(false);
	end
	
	self:AddToFrames(frameName)
	
	return f

end

--- ========================================================= ---
--- Rotate Texture
--- ========================================================= ---

function Sphere:RotateTexture(tex, degrees, xOrigin, yOrigin)

	local radians = -math.rad(degrees - 135 );

	local sin = math.sin(radians) * 0.71;
	local cos = math.cos(radians) * 0.71;

	tex:SetTexCoord(xOrigin - sin, yOrigin + cos, xOrigin + cos, yOrigin + sin, xOrigin - cos, yOrigin - sin, xOrigin + sin, yOrigin - cos);

end

--- ========================================================= ---
--- Create Sphere Frames
--- ========================================================= ---
function Sphere:CreateFrames()

	Sphere:CreateFrame("Frame", "SphereMainFrame", UIParent, "top", "", true)
	Sphere:CreateFrame("Button", "SphereBG", SphereMainFrame, 2, "Interface\\Addons\\Sphere\\art\\spherebg")
	Sphere:CreateFrame("Button", "SphereOuterFill", SphereMainFrame, 4, "Interface\\Addons\\Sphere\\art\\outerfill")
	Sphere:CreateFrame("Button", "SphereOuterCover", SphereMainFrame, 6, "Interface\\Addons\\Sphere\\art\\outercover")
	Sphere:CreateFrame("Button", "SphereOuterHalf", SphereMainFrame, 6, "Interface\\Addons\\Sphere\\art\\outerfill")
	Sphere:CreateFrame("Button", "SphereOuterBorder", SphereMainFrame, 8, "Interface\\Addons\\Sphere\\art\\outercircle")
	Sphere:CreateFrame("Button", "SphereOuterMarks", SphereMainFrame, 8, "Interface\\Addons\\Sphere\\art\\outermarks5")
	
	Sphere:RotateTexture(SphereOuterHalf:GetNormalTexture(), 180, 0.5, 0.5)
	SphereOuterHalf:Hide();
end

--- ========================================================= ---
--- Update Gauge
--- ========================================================= ---
function Sphere:UpdateGauge()
	local sPercent = 0
	local r, g, b, a
	
	if self.db.char.useCustom then
		sPercent, r, g, b, a  = Sphere[self.db.char.gaugeType](Sphere)
	else
		if self.db.char.gaugeType == "power" then
			local powerType = DogTag:Evaluate(self.db.char.gaugeUnit, "[TypePower]")
			if powerType == "" then powerType = "Mana" end
			r, g, b, a = unpack(self.db.char.colors[string.lower(powerType)])
		else
			r, g, b, a = unpack(self.db.char.colors[self.db.char.gaugeType])
		end
		
		if tonumber(DogTag:Evaluate(self.db.char.gaugeUnit, self.db.char.gaugeCustom)) then
			sPercent = tonumber(DogTag:Evaluate(self.db.char.gaugeUnit, self.db.char.gaugeCustom))
		else
			sPercent = 0
		end
	
		if self.db.char.animate then
			if (start == nil) then
				start = sPercent;
			end
			if (start > sPercent) then
				start = start - 1;
				if (start < sPercent) then
					start = sPercent;
				end
			else
				start = start + 1;
				if (start > sPercent) then
					start = sPercent;
				end
			end
			sPercent = start;
		end
	end

	if (sPercent > 100) then
		sPercent = 100;
	end

	if (sPercent <= 0) then
		SphereOuterHalf:Hide();
		SphereOuterHalf:SetID(0);
		SphereOuterCover:Show();
		SphereOuterFill:SetAlpha(0);
	end

	if (sPercent > 0) then
		if ((sPercent >= 50) and (SphereOuterHalf:GetID() ~= 1)) then
			SphereOuterHalf:Show();
			SphereOuterHalf:SetID(1);
			SphereOuterCover:Hide();
		elseif ((sPercent < 50) and (SphereOuterHalf:GetID() == 1)) then
			SphereOuterHalf:Hide();
			SphereOuterHalf:SetID(0);
			SphereOuterCover:Show();
		end
		SphereOuterFill:SetAlpha(1);
	else
		SphereOuterFill:SetAlpha(0);
	end
	
	SphereOuterFill:GetNormalTexture():SetVertexColor(r, g, b, a);
	SphereOuterHalf:GetNormalTexture():SetVertexColor(r, g, b, a);
	
	Sphere:RotateTexture(SphereOuterFill:GetNormalTexture(), math.floor(sPercent * 3.6), 0.5, 0.5)
	
end